@@ -28,7 +28,6 @@ import (
 	"github.com/moby/buildkit/util/testutil/echoserver"
 	"github.com/moby/buildkit/util/testutil/integration"
 	"github.com/moby/buildkit/util/testutil/workers"
-	digest "github.com/opencontainers/go-digest"
 	"github.com/pkg/errors"
 	"github.com/stretchr/testify/require"
 	"github.com/tonistiigi/fsutil"
@@ -201,7 +200,7 @@ func testWarnings(t *testing.T, sb integration.Sandbox) {
 				Data:     []byte("filedata"),
 			},
 			Range: []*pb.Range{
-				{Start: pb.Position{Line: 2}, End: pb.Position{Line: 4}},
+				{Start: &pb.Position{Line: 2}, End: &pb.Position{Line: 4}},
 			},
 			Detail: [][]byte{[]byte("this is detail"), []byte("and more detail")},
 			URL:    "https://example.com",
@@ -210,43 +209,14 @@ func testWarnings(t *testing.T, sb integration.Sandbox) {
 		return r, nil
 	}
 
-	status := make(chan *SolveStatus)
-	statusDone := make(chan struct{})
-	done := make(chan struct{})
+	wc := newWarningsCapture()
 
-	var warnings []*VertexWarning
-	vertexes := map[digest.Digest]struct{}{}
-
-	go func() {
-		defer close(statusDone)
-		for {
-			select {
-			case st, ok := <-status:
-				if !ok {
-					return
-				}
-				for _, s := range st.Vertexes {
-					vertexes[s.Digest] = struct{}{}
-				}
-				warnings = append(warnings, st.Warnings...)
-			case <-done:
-				return
-			}
-		}
-	}()
-
-	_, err = c.Build(ctx, SolveOpt{}, product, b, status)
+	_, err = c.Build(ctx, SolveOpt{}, product, b, wc.status)
 	require.NoError(t, err)
 
-	select {
-	case <-statusDone:
-	case <-time.After(10 * time.Second):
-		close(done)
-	}
+	warnings := wc.wait()
 
-	<-statusDone
-
-	require.Equal(t, 1, len(vertexes))
+	require.Equal(t, 1, len(wc.vertexes))
 	require.Equal(t, 1, len(warnings))
 
 	w := warnings[0]
@@ -257,7 +227,7 @@ func testWarnings(t *testing.T, sb integration.Sandbox) {
 	require.Equal(t, "and more detail", string(w.Detail[1]))
 	require.Equal(t, "https://example.com", w.URL)
 	require.Equal(t, 3, w.Level)
-	_, ok := vertexes[w.Vertex]
+	_, ok := wc.vertexes[w.Vertex]
 	require.True(t, ok)
 
 	require.Equal(t, "mydockerfile", w.SourceInfo.Filename)
@@ -339,7 +309,7 @@ func testUnknownBuildID(t *testing.T, sb integration.Sandbox) {
 	_, err = g.Ping(ctx, &gatewayapi.PingRequest{})
 	require.Error(t, err)
 	require.Contains(t, err.Error(), "no such job")
-	require.Equal(t, grpcerrors.Code(err), codes.NotFound)
+	require.Equal(t, codes.NotFound, grpcerrors.Code(err))
 }
 
 // testClientGatewayContainerCancelOnRelease is testing that all running
@@ -452,7 +422,6 @@ func testClientGatewayContainerExecPipe(t *testing.T, sb integration.Sandbox) {
 				Ref:       r.Ref,
 			}},
 		})
-
 		if err != nil {
 			return nil, err
 		}
@@ -462,13 +431,13 @@ func testClientGatewayContainerExecPipe(t *testing.T, sb integration.Sandbox) {
 			Args: []string{"sleep", "10"},
 		})
 		if err != nil {
-			ctr.Release(ctx)
+			ctr.Release(context.WithoutCancel(ctx))
 			return nil, err
 		}
 
 		defer func() {
 			// cancel pid1
-			ctr.Release(ctx)
+			ctr.Release(context.WithoutCancel(ctx))
 			pid1.Wait()
 		}()
 
@@ -483,7 +452,6 @@ func testClientGatewayContainerExecPipe(t *testing.T, sb integration.Sandbox) {
 			Stdin:  io.NopCloser(stdin2),
 			Stdout: stdout2,
 		})
-
 		if err != nil {
 			return nil, err
 		}
@@ -576,7 +544,6 @@ func testClientGatewayContainerPID1Fail(t *testing.T, sb integration.Sandbox) {
 				Ref:       r.Ref,
 			}},
 		})
-
 		if err != nil {
 			return nil, err
 		}
@@ -640,7 +607,6 @@ func testClientGatewayContainerPID1Exit(t *testing.T, sb integration.Sandbox) {
 				Ref:       r.Ref,
 			}},
 		})
-
 		if err != nil {
 			return nil, err
 		}
@@ -1221,7 +1187,7 @@ func testClientGatewayContainerExecTty(t *testing.T, sb integration.Sandbox) {
 	checkAllReleasable(t, c, sb, true)
 }
 
-// testClientGatewayContainerExecTty is testing the tty shuts down cleanly
+// testClientGatewayContainerCancelExecTty is testing the tty shuts down cleanly
 // on context.Cancel
 func testClientGatewayContainerCancelExecTty(t *testing.T, sb integration.Sandbox) {
 	requiresLinux(t)
@@ -1270,7 +1236,7 @@ func testClientGatewayContainerCancelExecTty(t *testing.T, sb integration.Sandbo
 		defer ctr.Release(ctx)
 
 		execCtx, cancel := context.WithCancelCause(ctx)
-		defer cancel(errors.WithStack(context.Canceled))
+		defer func() { cancel(errors.WithStack(context.Canceled)) }()
 
 		prompt := newTestPrompt(execCtx, t, inputW, output)
 		pid2, err := ctr.Start(execCtx, client.StartRequest{
@@ -1462,28 +1428,32 @@ func testClientGatewayExecError(t *testing.T, sb integration.Sandbox) {
 			llb.Image("busybox:latest").Run(
 				llb.Shlexf(`sh -c "echo %s > /data && fail"`, id),
 			).Root(),
-			1, []string{"/data"},
+			1,
+			[]string{"/data"},
 		}, {
 			"rootfs and readwrite scratch mount",
 			llb.Image("busybox:latest").Run(
 				llb.Shlexf(`sh -c "echo %s > /data && echo %s > /rw/data && fail"`, id, id),
 				llb.AddMount("/rw", llb.Scratch()),
 			).Root(),
-			2, []string{"/data", "/rw/data"},
+			2,
+			[]string{"/data", "/rw/data"},
 		}, {
 			"rootfs and readwrite mount",
 			llb.Image("busybox:latest").Run(
 				llb.Shlexf(`sh -c "echo %s > /data && echo %s > /rw/data && fail"`, id, id),
 				llb.AddMount("/rw", llb.Scratch().File(llb.Mkfile("foo", 0700, []byte(id)))),
 			).Root(),
-			2, []string{"/data", "/rw/data", "/rw/foo"},
+			2,
+			[]string{"/data", "/rw/data", "/rw/foo"},
 		}, {
 			"rootfs and readonly scratch mount",
 			llb.Image("busybox:latest").Run(
 				llb.Shlexf(`sh -c "echo %s > /data && echo %s > /readonly/foo"`, id, id),
 				llb.AddMount("/readonly", llb.Scratch(), llb.Readonly),
 			).Root(),
-			2, []string{"/data"},
+			2,
+			[]string{"/data"},
 		}, {
 			"rootfs and readwrite force no output mount",
 			llb.Image("busybox:latest").Run(
@@ -1494,7 +1464,8 @@ func testClientGatewayExecError(t *testing.T, sb integration.Sandbox) {
 					llb.ForceNoOutput,
 				),
 			).Root(),
-			2, []string{"/data", "/rw/data", "/rw/foo"},
+			2,
+			[]string{"/data", "/rw/data", "/rw/foo"},
 		}}
 
 		for _, tt := range tests {
@@ -1877,7 +1848,7 @@ func testClientGatewayContainerSecurityMode(t *testing.T, sb integration.Sandbox
 				0x00000000a80425fb=cap_chown,cap_dac_override,cap_fowner,cap_fsetid,cap_kill,cap_setgid,cap_setuid,cap_setpcap,
 				cap_net_bind_service,cap_net_raw,cap_sys_chroot,cap_mknod,cap_audit_write,cap_setfcap
 			*/
-			require.EqualValues(t, 0xa80425fb, caps)
+			require.Equal(t, uint64(0xa80425fb), caps)
 		}
 		allowedEntitlements = []entitlements.Entitlement{}
 		if expectFail {
@@ -1895,7 +1866,7 @@ func testClientGatewayContainerSecurityMode(t *testing.T, sb integration.Sandbox
 			*/
 
 			// require that _at least_ minimum capabilities are granted
-			require.EqualValues(t, 0x3fffffffff, caps&0x3fffffffff)
+			require.Equal(t, uint64(0x3fffffffff), caps&0x3fffffffff)
 		}
 		mode = llb.SecurityModeInsecure
 		allowedEntitlements = []entitlements.Entitlement{entitlements.EntitlementSecurityInsecure}
@@ -1926,7 +1897,6 @@ func testClientGatewayContainerSecurityMode(t *testing.T, sb integration.Sandbox
 				Ref:       r.Ref,
 			}},
 		})
-
 		if err != nil {
 			return nil, err
 		}
@@ -2018,7 +1988,6 @@ func testClientGatewayContainerExtraHosts(t *testing.T, sb integration.Sandbox)
 				IP:   "169.254.11.22",
 			}},
 		})
-
 		if err != nil {
 			return nil, err
 		}
@@ -2118,7 +2087,6 @@ func testClientGatewayContainerHostNetworking(t *testing.T, sb integration.Sandb
 			}},
 			NetMode: netMode,
 		})
-
 		if err != nil {
 			return nil, err
 		}
