@@ -8,6 +8,7 @@ import (
 	"github.com/moby/buildkit/solver/pb"
 	digest "github.com/opencontainers/go-digest"
 	"github.com/stretchr/testify/require"
+	"golang.org/x/sync/errgroup"
 )
 
 func TestFileMkdir(t *testing.T) {
@@ -26,8 +27,8 @@ func TestFileMkdir(t *testing.T) {
 	require.Equal(t, m[dgst], arr[1])
 
 	f := arr[1].Op.(*pb.Op_File).File
-	require.Equal(t, len(arr[1].Inputs), 1)
-	require.Equal(t, m[arr[1].Inputs[0].Digest], arr[0])
+	require.Equal(t, 1, len(arr[1].Inputs))
+	require.Equal(t, arr[0], m[arr[1].Inputs[0].Digest])
 	require.Equal(t, 0, int(arr[1].Inputs[0].Index))
 
 	require.Equal(t, 1, len(f.Actions))
@@ -60,7 +61,7 @@ func TestFileMkdirChain(t *testing.T) {
 	require.Equal(t, m[dgst], arr[1])
 
 	f := arr[1].Op.(*pb.Op_File).File
-	require.Equal(t, len(arr[1].Inputs), 1)
+	require.Equal(t, 1, len(arr[1].Inputs))
 	require.Equal(t, m[arr[1].Inputs[0].Digest], arr[0])
 	require.Equal(t, 0, int(arr[1].Inputs[0].Index))
 
@@ -113,7 +114,7 @@ func TestFileMkdirMkfile(t *testing.T) {
 	require.Equal(t, m[dgst], arr[0])
 
 	f := arr[0].Op.(*pb.Op_File).File
-	require.Equal(t, len(arr[1].Inputs), 1)
+	require.Equal(t, 1, len(arr[1].Inputs))
 	require.Equal(t, m[arr[1].Inputs[0].Digest], arr[0])
 	require.Equal(t, 0, int(arr[1].Inputs[0].Index))
 
@@ -159,7 +160,7 @@ func TestFileMkfile(t *testing.T) {
 	require.Equal(t, m[dgst], arr[1])
 
 	f := arr[1].Op.(*pb.Op_File).File
-	require.Equal(t, len(arr[1].Inputs), 1)
+	require.Equal(t, 1, len(arr[1].Inputs))
 	require.Equal(t, m[arr[1].Inputs[0].Digest], arr[0])
 	require.Equal(t, 0, int(arr[1].Inputs[0].Index))
 
@@ -194,7 +195,7 @@ func TestFileRm(t *testing.T) {
 	require.Equal(t, m[dgst], arr[1])
 
 	f := arr[1].Op.(*pb.Op_File).File
-	require.Equal(t, len(arr[1].Inputs), 1)
+	require.Equal(t, 1, len(arr[1].Inputs))
 	require.Equal(t, m[arr[1].Inputs[0].Digest], arr[0])
 	require.Equal(t, 0, int(arr[1].Inputs[0].Index))
 
@@ -235,7 +236,7 @@ func TestFileSimpleChains(t *testing.T) {
 	require.Equal(t, m[dgst], arr[2])
 
 	f := arr[2].Op.(*pb.Op_File).File
-	require.Equal(t, len(arr[2].Inputs), 1)
+	require.Equal(t, 1, len(arr[2].Inputs))
 	require.Equal(t, m[arr[2].Inputs[0].Digest], arr[1])
 	require.Equal(t, 0, int(arr[2].Inputs[0].Index))
 	require.Equal(t, 2, len(f.Actions))
@@ -257,7 +258,7 @@ func TestFileSimpleChains(t *testing.T) {
 	require.Equal(t, "/abc", mkfile.Path)
 
 	f = arr[1].Op.(*pb.Op_File).File
-	require.Equal(t, len(arr[1].Inputs), 1)
+	require.Equal(t, 1, len(arr[1].Inputs))
 	require.Equal(t, m[arr[1].Inputs[0].Digest], arr[0])
 	require.Equal(t, 0, int(arr[1].Inputs[0].Index))
 	require.Equal(t, 3, len(f.Actions))
@@ -513,7 +514,7 @@ func TestFileOwner(t *testing.T) {
 	require.Equal(t, m[dgst], arr[1])
 
 	f := arr[1].Op.(*pb.Op_File).File
-	require.Equal(t, len(arr[1].Inputs), 1)
+	require.Equal(t, 1, len(arr[1].Inputs))
 	require.Equal(t, m[arr[1].Inputs[0].Digest], arr[0])
 	require.Equal(t, 0, int(arr[1].Inputs[0].Index))
 
@@ -610,7 +611,7 @@ func TestFileCopyOwner(t *testing.T) {
 	require.Equal(t, m[dgst], arr[1])
 
 	f := arr[1].Op.(*pb.Op_File).File
-	require.Equal(t, len(arr[1].Inputs), 1)
+	require.Equal(t, 1, len(arr[1].Inputs))
 	require.Equal(t, m[arr[1].Inputs[0].Digest], arr[0])
 	require.Equal(t, 0, int(arr[1].Inputs[0].Index))
 
@@ -673,7 +674,7 @@ func TestFileCreatedTime(t *testing.T) {
 	require.Equal(t, m[dgst], arr[1])
 
 	f := arr[1].Op.(*pb.Op_File).File
-	require.Equal(t, len(arr[1].Inputs), 1)
+	require.Equal(t, 1, len(arr[1].Inputs))
 	require.Equal(t, m[arr[1].Inputs[0].Digest], arr[0])
 	require.Equal(t, 0, int(arr[1].Inputs[0].Index))
 
@@ -692,27 +693,60 @@ func TestFileCreatedTime(t *testing.T) {
 	require.Equal(t, dt3.UnixNano(), copy.Timestamp)
 }
 
-func parseDef(t *testing.T, def [][]byte) (map[digest.Digest]pb.Op, []pb.Op) {
-	m := map[digest.Digest]pb.Op{}
-	arr := make([]pb.Op, 0, len(def))
+func parseDef(t *testing.T, def [][]byte) (map[string]*pb.Op, []*pb.Op) {
+	m := map[string]*pb.Op{}
+	arr := make([]*pb.Op, 0, len(def))
 
 	for _, dt := range def {
 		var op pb.Op
-		err := (&op).Unmarshal(dt)
+		err := op.Unmarshal(dt)
 		require.NoError(t, err)
 		dgst := digest.FromBytes(dt)
-		m[dgst] = op
-		arr = append(arr, op)
+		m[string(dgst)] = &op
+		arr = append(arr, &op)
 		// fmt.Printf(":: %T %+v\n", op.Op, op)
 	}
 
 	return m, arr
 }
 
-func last(t *testing.T, arr []pb.Op) (digest.Digest, int) {
-	require.True(t, len(arr) > 1)
+func last(t *testing.T, arr []*pb.Op) (string, int) {
+	require.Greater(t, len(arr), 1)
 
 	op := arr[len(arr)-1]
 	require.Equal(t, 1, len(op.Inputs))
 	return op.Inputs[0].Digest, int(op.Inputs[0].Index)
 }
+
+func TestFileOpMarshalConsistency(t *testing.T) {
+	var prevDef [][]byte
+
+	f1 := Scratch().File(Mkfile("/tmp", 0644, []byte("tmp 1")))
+	f2 := Scratch().File(Mkfile("/a", 0644, []byte("tmp 2")))
+	st := Image("foo").Dir("/tmp").
+		File(Copy(f1, "/foo", "/bar")).
+		File(Copy(f2, "/a", "/b"))
+
+	for i := 0; i < 100; i++ {
+		def, err := st.Marshal(context.TODO())
+		require.NoError(t, err)
+
+		if prevDef != nil {
+			require.Equal(t, def.Def, prevDef)
+		}
+
+		prevDef = def.Def
+	}
+}
+
+func TestParallelMarshal(t *testing.T) {
+	st := Scratch().File(Mkfile("/tmp", 0644, []byte("tmp 1")))
+	eg, ctx := errgroup.WithContext(context.Background())
+	for i := 0; i < 100; i++ {
+		eg.Go(func() error {
+			_, err := st.Marshal(ctx)
+			return err
+		})
+	}
+	require.NoError(t, eg.Wait())
+}
