@@ -1,5 +1,4 @@
 //go:build windows
-// +build windows
 
 package main
 
@@ -17,7 +16,7 @@ import (
 
 const socketScheme = "npipe://"
 
-func listenFD(addr string, tlsConfig *tls.Config) (net.Listener, error) {
+func listenFD(_ string, _ *tls.Config) (net.Listener, error) {
 	return nil, errors.New("listening server on fd not supported on windows")
 }
 
