@@ -19,8 +19,8 @@ import (
 	"github.com/containerd/containerd/content"
 	"github.com/containerd/containerd/content/local"
 	"github.com/containerd/containerd/content/proxy"
-	"github.com/containerd/containerd/platforms"
 	"github.com/containerd/continuity/fs/fstest"
+	"github.com/containerd/platforms"
 	intoto "github.com/in-toto/in-toto-golang/in_toto"
 	provenanceCommon "github.com/in-toto/in-toto-golang/in_toto/slsa_provenance/common"
 	controlapi "github.com/moby/buildkit/api/services/control"
@@ -36,6 +36,7 @@ import (
 	"github.com/moby/buildkit/util/testutil"
 	"github.com/moby/buildkit/util/testutil/integration"
 	"github.com/moby/buildkit/util/testutil/workers"
+	digest "github.com/opencontainers/go-digest"
 	ocispecs "github.com/opencontainers/image-spec/specs-go/v1"
 	"github.com/pkg/errors"
 	"github.com/stretchr/testify/require"
@@ -43,7 +44,6 @@ import (
 )
 
 func testProvenanceAttestation(t *testing.T, sb integration.Sandbox) {
-	integration.SkipOnPlatform(t, "windows")
 	workers.CheckFeatureCompat(t, sb, workers.FeatureDirectPush, workers.FeatureProvenance)
 	ctx := sb.Context()
 
@@ -59,10 +59,18 @@ func testProvenanceAttestation(t *testing.T, sb integration.Sandbox) {
 
 	f := getFrontend(t, sb)
 
-	dockerfile := []byte(`
+	dockerfile := []byte(integration.UnixOrWindows(
+		`
 FROM busybox:latest
 RUN echo "ok" > /foo
-`)
+`,
+		`
+FROM nanoserver
+USER ContainerAdministrator
+RUN echo ok> /foo
+`,
+	))
+
 	dir := integration.Tmpdir(
 		t,
 		fstest.CreateFile("Dockerfile", dockerfile, 0600),
@@ -115,12 +123,14 @@ RUN echo "ok" > /foo
 
 			img := imgs.Find(platforms.Format(platforms.Normalize(platforms.DefaultSpec())))
 			require.NotNil(t, img)
-			require.Equal(t, []byte("ok\n"), img.Layers[1]["foo"].Data)
+			outFile := integration.UnixOrWindows("foo", "Files/foo")
+			expectedFileData := integration.UnixOrWindows([]byte("ok\n"), []byte("ok\r\n"))
+			require.Equal(t, expectedFileData, img.Layers[1][outFile].Data)
 
 			att := imgs.Find("unknown/unknown")
 			require.NotNil(t, att)
-			require.Equal(t, att.Desc.Annotations["vnd.docker.reference.digest"], string(img.Desc.Digest))
-			require.Equal(t, att.Desc.Annotations["vnd.docker.reference.type"], "attestation-manifest")
+			require.Equal(t, string(img.Desc.Digest), att.Desc.Annotations["vnd.docker.reference.digest"])
+			require.Equal(t, "attestation-manifest", att.Desc.Annotations["vnd.docker.reference.type"])
 			var attest intoto.Statement
 			require.NoError(t, json.Unmarshal(att.LayersRaw[0], &attest))
 			require.Equal(t, "https://in-toto.io/Statement/v0.1", attest.Type)
@@ -179,7 +189,9 @@ RUN echo "ok" > /foo
 				require.Equal(t, "https://xxxxx:xxxxx@example.invalid/foo.html", args["context:foo"])
 			}
 
-			expectedBase := "pkg:docker/busybox@latest?platform=" + url.PathEscape(platforms.Format(platforms.Normalize(platforms.DefaultSpec())))
+			expectedBaseImage := integration.UnixOrWindows("busybox", "nanoserver")
+			escapedPlatform := url.PathEscape(platforms.Format(platforms.Normalize(platforms.DefaultSpec())))
+			expectedBase := fmt.Sprintf("pkg:docker/%s@latest?platform=%s", expectedBaseImage, escapedPlatform)
 			if isGateway {
 				require.Equal(t, 2, len(pred.Materials), "%+v", pred.Materials)
 				require.Contains(t, pred.Materials[0].URI, "docker/buildkit_test")
@@ -204,9 +216,9 @@ RUN echo "ok" > /foo
 			require.Equal(t, "dockerfile", pred.Invocation.Parameters.Locals[1].Name)
 
 			require.NotNil(t, pred.Metadata.BuildFinishedOn)
-			require.True(t, time.Since(*pred.Metadata.BuildFinishedOn) < 5*time.Minute)
+			require.Less(t, time.Since(*pred.Metadata.BuildFinishedOn), 5*time.Minute)
 			require.NotNil(t, pred.Metadata.BuildStartedOn)
-			require.True(t, time.Since(*pred.Metadata.BuildStartedOn) < 5*time.Minute)
+			require.Less(t, time.Since(*pred.Metadata.BuildStartedOn), 5*time.Minute)
 			require.True(t, pred.Metadata.BuildStartedOn.Before(*pred.Metadata.BuildFinishedOn))
 
 			require.True(t, pred.Metadata.Completeness.Environment)
@@ -276,7 +288,7 @@ COPY myapp.Dockerfile /
 	expectedGitSHA, err := cmd.Output()
 	require.NoError(t, err)
 
-	server := httptest.NewServer(http.FileServer(http.Dir(filepath.Join(dir.Name))))
+	server := httptest.NewServer(http.FileServer(http.Dir(filepath.Clean(dir.Name))))
 	defer server.Close()
 
 	target := registry + "/buildkit/testwithprovenance:git"
@@ -316,8 +328,8 @@ COPY myapp.Dockerfile /
 
 	att := imgs.Find("unknown/unknown")
 	require.NotNil(t, att)
-	require.Equal(t, att.Desc.Annotations["vnd.docker.reference.digest"], string(img.Desc.Digest))
-	require.Equal(t, att.Desc.Annotations["vnd.docker.reference.type"], "attestation-manifest")
+	require.Equal(t, string(img.Desc.Digest), att.Desc.Annotations["vnd.docker.reference.digest"])
+	require.Equal(t, "attestation-manifest", att.Desc.Annotations["vnd.docker.reference.type"])
 	var attest intoto.Statement
 	require.NoError(t, json.Unmarshal(att.LayersRaw[0], &attest))
 	require.Equal(t, "https://in-toto.io/Statement/v0.1", attest.Type)
@@ -453,7 +465,7 @@ RUN echo "ok-$TARGETARCH" > /foo
 
 		att := imgs.FindAttestation(p)
 		require.NotNil(t, att)
-		require.Equal(t, att.Desc.Annotations["vnd.docker.reference.type"], "attestation-manifest")
+		require.Equal(t, "attestation-manifest", att.Desc.Annotations["vnd.docker.reference.type"])
 		var attest intoto.Statement
 		require.NoError(t, json.Unmarshal(att.LayersRaw[0], &attest))
 		require.Equal(t, "https://in-toto.io/Statement/v0.1", attest.Type)
@@ -635,7 +647,7 @@ func testClientFrontendProvenance(t *testing.T, sb integration.Sandbox) {
 
 	att := imgs.FindAttestation("linux/arm64")
 	require.NotNil(t, att)
-	require.Equal(t, att.Desc.Annotations["vnd.docker.reference.type"], "attestation-manifest")
+	require.Equal(t, "attestation-manifest", att.Desc.Annotations["vnd.docker.reference.type"])
 	var attest intoto.Statement
 	require.NoError(t, json.Unmarshal(att.LayersRaw[0], &attest))
 	require.Equal(t, "https://in-toto.io/Statement/v0.1", attest.Type)
@@ -668,7 +680,7 @@ func testClientFrontendProvenance(t *testing.T, sb integration.Sandbox) {
 
 	att = imgs.FindAttestation("linux/amd64")
 	require.NotNil(t, att)
-	require.Equal(t, att.Desc.Annotations["vnd.docker.reference.type"], "attestation-manifest")
+	require.Equal(t, "attestation-manifest", att.Desc.Annotations["vnd.docker.reference.type"])
 	attest = intoto.Statement{}
 	require.NoError(t, json.Unmarshal(att.LayersRaw[0], &attest))
 	require.Equal(t, "https://in-toto.io/Statement/v0.1", attest.Type)
@@ -693,7 +705,6 @@ func testClientFrontendProvenance(t *testing.T, sb integration.Sandbox) {
 }
 
 func testClientLLBProvenance(t *testing.T, sb integration.Sandbox) {
-	integration.SkipOnPlatform(t, "windows")
 	workers.CheckFeatureCompat(t, sb, workers.FeatureDirectPush, workers.FeatureProvenance)
 	ctx := sb.Context()
 
@@ -736,7 +747,7 @@ func testClientLLBProvenance(t *testing.T, sb integration.Sandbox) {
 			return nil, err
 		}
 
-		st = llb.Image("alpine").File(llb.Mkfile("/foo", 0600, dt))
+		st = llb.Image(integration.UnixOrWindows("alpine", "nanoserver")).File(llb.Mkfile("/foo", 0600, dt))
 		def, err = st.Marshal(ctx)
 		if err != nil {
 			return nil, err
@@ -776,12 +787,13 @@ func testClientLLBProvenance(t *testing.T, sb integration.Sandbox) {
 	nativePlatform := platforms.Format(platforms.Normalize(platforms.DefaultSpec()))
 
 	img := imgs.Find(nativePlatform)
+	fileName := integration.UnixOrWindows("foo", "Files/foo")
 	require.NotNil(t, img)
-	require.Contains(t, string(img.Layers[1]["foo"].Data), "The Moby Project")
+	require.Contains(t, string(img.Layers[1][fileName].Data), "The Moby Project")
 
 	att := imgs.FindAttestation(nativePlatform)
 	require.NotNil(t, att)
-	require.Equal(t, att.Desc.Annotations["vnd.docker.reference.type"], "attestation-manifest")
+	require.Equal(t, "attestation-manifest", att.Desc.Annotations["vnd.docker.reference.type"])
 	var attest intoto.Statement
 	require.NoError(t, json.Unmarshal(att.LayersRaw[0], &attest))
 	require.Equal(t, "https://in-toto.io/Statement/v0.1", attest.Type)
@@ -803,7 +815,7 @@ func testClientLLBProvenance(t *testing.T, sb integration.Sandbox) {
 	require.Equal(t, 0, len(pred.Invocation.Parameters.Locals))
 
 	require.Equal(t, 2, len(pred.Materials), "%+v", pred.Materials)
-	require.Contains(t, pred.Materials[0].URI, "docker/alpine")
+	require.Contains(t, pred.Materials[0].URI, integration.UnixOrWindows("docker/alpine", "docker/nanoserver"))
 	require.Contains(t, pred.Materials[1].URI, "README.md")
 }
 
@@ -1020,7 +1032,6 @@ EOF
 }
 
 func testNilProvenance(t *testing.T, sb integration.Sandbox) {
-	integration.SkipOnPlatform(t, "windows")
 	workers.CheckFeatureCompat(t, sb, workers.FeatureProvenance)
 	ctx := sb.Context()
 
@@ -1030,10 +1041,16 @@ func testNilProvenance(t *testing.T, sb integration.Sandbox) {
 
 	f := getFrontend(t, sb)
 
-	dockerfile := []byte(`
+	dockerfile := []byte(integration.UnixOrWindows(
+		`
 FROM scratch
 ENV FOO=bar
-`)
+`,
+		`
+FROM scratch
+ENV FOO=bar
+`,
+	))
 	dir := integration.Tmpdir(
 		t,
 		fstest.CreateFile("Dockerfile", dockerfile, 0600),
@@ -1058,7 +1075,6 @@ ENV FOO=bar
 
 // https://github.com/moby/buildkit/issues/3562
 func testDuplicatePlatformProvenance(t *testing.T, sb integration.Sandbox) {
-	integration.SkipOnPlatform(t, "windows")
 	workers.CheckFeatureCompat(t, sb, workers.FeatureProvenance)
 	ctx := sb.Context()
 
@@ -1068,12 +1084,17 @@ func testDuplicatePlatformProvenance(t *testing.T, sb integration.Sandbox) {
 
 	f := getFrontend(t, sb)
 
-	dockerfile := []byte(`FROM alpine`)
+	dockerfile := []byte(
+		`
+FROM alpine as base-linux
+FROM nanoserver as base-windows
+FROM base-$TARGETOS
+`,
+	)
 	dir := integration.Tmpdir(
 		t,
 		fstest.CreateFile("Dockerfile", dockerfile, 0600),
 	)
-
 	_, err = f.Solve(sb.Context(), c, client.SolveOpt{
 		FrontendAttrs: map[string]string{
 			"attest:provenance": "mode=max",
@@ -1089,13 +1110,19 @@ func testDuplicatePlatformProvenance(t *testing.T, sb integration.Sandbox) {
 
 // https://github.com/moby/buildkit/issues/3928
 func testDockerIgnoreMissingProvenance(t *testing.T, sb integration.Sandbox) {
-	integration.SkipOnPlatform(t, "windows")
 	workers.CheckFeatureCompat(t, sb, workers.FeatureProvenance)
 	c, err := client.New(sb.Context(), sb.Address())
 	require.NoError(t, err)
 	defer c.Close()
 
-	dockerfile := []byte(`FROM alpine`)
+	dockerfile := []byte(integration.UnixOrWindows(
+		`
+FROM alpine
+`,
+		`
+FROM nanoserver
+`,
+	))
 	dirDockerfile := integration.Tmpdir(
 		t,
 		fstest.CreateFile("Dockerfile", dockerfile, 0600),
@@ -1250,20 +1277,22 @@ ADD bar bar`)
 }
 
 func testFrontendDeduplicateSources(t *testing.T, sb integration.Sandbox) {
-	integration.SkipOnPlatform(t, "windows")
 	ctx := sb.Context()
 
 	c, err := client.New(ctx, sb.Address())
 	require.NoError(t, err)
 	defer c.Close()
 
-	dockerfile := []byte(`
-FROM scratch as base
+	dockerfile := []byte(fmt.Sprintf(
+		`
+FROM %s as base
 COPY foo foo2
 
 FROM linked
 COPY bar bar2
-`)
+`,
+		integration.UnixOrWindows("scratch", "nanoserver")),
+	)
 
 	dir := integration.Tmpdir(
 		t,
@@ -1369,6 +1398,7 @@ COPY bar bar2
 			break
 		}
 		require.Equal(t, ref, ev.Record.Ref)
+		require.Len(t, ev.Record.Exporters, 1)
 
 		for _, prov := range ev.Record.Result.Attestations {
 			if len(prov.Annotations) == 0 || prov.Annotations["in-toto.io/predicate-type"] != "https://slsa.dev/provenance/v0.2" {
@@ -1378,14 +1408,14 @@ COPY bar bar2
 
 			provDt, err = content.ReadBlob(ctx, store, ocispecs.Descriptor{
 				MediaType: prov.MediaType,
-				Digest:    prov.Digest,
-				Size:      prov.Size_,
+				Digest:    digest.Digest(prov.Digest),
+				Size:      prov.Size,
 			})
 			require.NoError(t, err)
 		}
 	}
 
-	require.NotEqual(t, len(provDt), 0)
+	require.NotEqual(t, 0, len(provDt))
 
 	var pred provenancetypes.ProvenancePredicate
 	require.NoError(t, json.Unmarshal(provDt, &pred))
