@@ -11,6 +11,11 @@ func TestParseScopes(t *testing.T) {
 		input    []string
 		expected scopes
 	}{
+		{
+			name:     "InvalidScope",
+			input:    []string{""},
+			expected: nil,
+		},
 		{
 			name: "SeparateStrings",
 			input: []string{
